     //////////////////////////////////////////////////////////////////////
     // The contents of this file are subject to the Mozilla Public License
     // Version 1.1 (the "License"); you may not use this file except in
     // compliance with the License.  You may obtain a copy of the License
     // at http://www.mozilla.org/MPL/
     //
     // Software distributed under the License is distributed on an "AS IS"
     // basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
     // See the License for the specific language governing rights and
     // limitations under the License.
     //
     // The Original Code is
     // "com.orbs.open1.mpl.app.retirement.RetirementServlet",
     // created 1999-08-15.
     //
     // The Initial Developer of the Original Code is David Wallace Croft
     // (croft@alumni.caltech.edu, http://www.alumni.caltech.edu/~croft/).
     // Portions created by David Wallace Croft are
     // Copyright (C) 1999 David Wallace Croft.  All Rights Reserved.
     //
     // Contributor(s):
     //////////////////////////////////////////////////////////////////////

     package com.orbs.open1.mpl.app.retirement;

     import java.io.*;
     import java.text.NumberFormat;

     import javax.servlet.*;
     import javax.servlet.http.*;

     import com.orbs.open1.mpl.io.FileLib;
     import com.orbs.open1.mpl.math.FinanceLib;
     import com.orbs.open1.mpl.text.ParseLib;

     /*********************************************************************
     * Retirement Calculator servlet.
     *
     * @version
     *   1999-08-15
     * @author
     *   <A HREF="http://www.alumni.caltech.edu/~croft/">David W. Croft</A>
     *********************************************************************/

     public final class  RetirementServlet extends HttpServlet
     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////
     {

     //////////////////////////////////////////////////////////////////////
     // Static variables
     //////////////////////////////////////////////////////////////////////

     private static final String  TITLE   = "Retirement Calculator";

     private static final String  VERSION = "1999-08-15";

     private static final String  DEFAULT_HEADER
       = "<HTML>\n<HEAD>\n<TITLE>" + TITLE + "</TITLE>\n</HEAD>\n"
       + "<BODY>\n<CENTER>\n<H1>" + TITLE + "</H1>\n"
       + "Version " + VERSION + "<BR>\n"
       + "<A HREF=\"http://www.alumni.caltech.edu/~croft/\">"
       + "David Wallace Croft</A>\n"
       + "<P>\n";

     private static final String  DEFAULT_FOOTER
       = "</CENTER>\n</BODY>\n</HTML>";

     private static final String  DEFAULT_HEADER_ALIAS_PATH
       = "/projects/retirement/header.html";

     private static final String  DEFAULT_FOOTER_ALIAS_PATH
       = "/projects/retirement/footer.html";

     private static final String [ ]  PARAM_TEXT = {
       "Desired annual retirement income (present value, after taxes)",
       "Estimated annual interest on retirement savings",
       "Estimated annual inflation",
       "Estimated tax rate during retirement",
       "Estimated annual growth on investments (tax-deferred)",
       "Number of years until retirement" };

     private static final String [ ]  PARAM_NAMES = {
       "rIncome",
       "rInterest",
       "rInflation",
       "rTaxRate",
       "iInterest",
       "iYears" };

     private static final double [ ]  PARAM_DEFAULT_VALUES = {
       30000.0,
       0.06,
       0.02,
       0.28,
       0.10,
       40.0 };

     private static final NumberFormat  currencyNumberFormat
       = NumberFormat.getCurrencyInstance ( );

     //////////////////////////////////////////////////////////////////////
     // Instance variables
     //////////////////////////////////////////////////////////////////////

     private String  htmlHeader;

     private String  htmlFooter;

     //////////////////////////////////////////////////////////////////////
     // Public static methods
     //////////////////////////////////////////////////////////////////////

     public static final double  calculateRequiredAnnualInvestment (
       double  desiredSavingsInterestIncome,
       double  savingsInterestRate,
       double  inflationRate,
       double  taxRate,
       double  investmentInterestRate,
       double  yearsOfSaving )
     //////////////////////////////////////////////////////////////////////
     {
       double  savings
         = desiredSavingsInterestIncome * ( 1.0 + inflationRate )
         / ( savingsInterestRate * ( 1.0 - taxRate ) - inflationRate );

       double  futureValueSavings
         = savings * Math.pow ( 1.0 + inflationRate, yearsOfSaving );

       double  annualSavings = FinanceLib.annualSavingsNeeded (
         futureValueSavings, investmentInterestRate, yearsOfSaving );

       return annualSavings;
     }

     //////////////////////////////////////////////////////////////////////
     // HttpServlet methods
     //////////////////////////////////////////////////////////////////////

     public String  getServletInfo ( ) { return TITLE; }

     public void  init ( ServletConfig  servletConfig )
       throws ServletException
     //////////////////////////////////////////////////////////////////////
     {
       try
       {
         super.init ( servletConfig );

         ServletContext  servletContext = getServletContext ( );

         String  headerRealPath
           = servletContext.getRealPath ( DEFAULT_HEADER_ALIAS_PATH );

         String  footerRealPath
           = servletContext.getRealPath ( DEFAULT_FOOTER_ALIAS_PATH );

         try
         {
           htmlHeader = FileLib.loadTextFile ( headerRealPath );
         }
         catch ( Exception  ex )
         {
           htmlHeader = "<! Unable to load \"" + headerRealPath + "\">\n"
             + "<! " + ex + " >\n"
             + DEFAULT_HEADER;
         }

         try
         {
           htmlFooter = FileLib.loadTextFile ( footerRealPath );
         }
         catch ( Exception  ex )
         {
           htmlFooter = "<! Unable to load \"" + footerRealPath + "\">\n"
             + "<! " + ex + " >\n"
             + DEFAULT_FOOTER;
         }
       }
       catch ( Exception  ex )
       {
         log ( "init() Exception", ex );
       }
     }

     public void  doGet (
       HttpServletRequest   req,
       HttpServletResponse  res )
       throws ServletException, IOException
     //////////////////////////////////////////////////////////////////////
     {
       doPost ( req, res );
     }

     public void  doPost (
       HttpServletRequest   req,
       HttpServletResponse  res )
       throws ServletException, IOException
     //////////////////////////////////////////////////////////////////////
     {
       res.setContentType ( "text/html" );

       PrintStream  out = new PrintStream ( res.getOutputStream ( ) );

       try
       {
         out.println ( htmlHeader );

         double [ ]  paramDoubles = new double [ PARAM_NAMES.length ];

         for ( int  i = 0; i < paramDoubles.length; i++ )
         {
           String [ ]  paramValues
             = req.getParameterValues ( PARAM_NAMES [ i ] );

           String  paramValue = null;

           if ( ( paramValues != null ) && ( paramValues.length > 0 ) )
           {
             paramValue = paramValues [ 0 ];
           }

           paramDoubles [ i ] = ParseLib.parseDouble (
             paramValue, PARAM_DEFAULT_VALUES [ i ] );
         }

         out.println ( "<FORM ACTION=\""
           + req.getRequestURI ( ) + "\" METHOD=GET>\n" );

         out.println ( "<TABLE>" );

         for ( int  i = 0; i < PARAM_TEXT.length; i++ )
         {
           out.println ( "<TR>" );

           out.println ( "  <TD>" + PARAM_TEXT [ i ] + "</TD>" );

           out.println ( "  <TD><INPUT TYPE=TEXT NAME=\""
             + PARAM_NAMES [ i ]
             + "\" SIZE=10 VALUE=\""
             + paramDoubles [ i ]
             + "\"></TD>" );

           out.println ( "</TR>" );
         }

         out.println ( "</TABLE>" );

         double  annualSavings = calculateRequiredAnnualInvestment (
           paramDoubles [ 0 ],
           paramDoubles [ 1 ],
           paramDoubles [ 2 ],
           paramDoubles [ 3 ],
           paramDoubles [ 4 ],
           paramDoubles [ 5 ] );

         out.println ( "You would need to invest "
           + currencyNumberFormat.format ( annualSavings ) + " a year." );

         out.println ( "<P>" );

         out.println ( "<INPUT TYPE=\"SUBMIT\" VALUE=\"Recalculate\">" );

         out.println ( "</FORM>\n" );

         out.println ( "<P>" );

         showTable ( out, annualSavings,
           paramDoubles [ 0 ], paramDoubles [ 1 ],
           paramDoubles [ 2 ], paramDoubles [ 3 ],
           paramDoubles [ 4 ], paramDoubles [ 5 ] );
       }
       catch ( Exception  ex )
       {
         ex.printStackTrace ( );

         out.println ( "A processing error occurred.  Please try again." );
       }
       catch ( Throwable  t )
       {
         t.printStackTrace ( );
       }
       finally
       {
         try
         {
           out.println ( htmlFooter );

           out.flush ( );

           out.close ( );
         }
         catch ( Exception  ex ) { }
       }
     }

     public void  destroy ( )
     //////////////////////////////////////////////////////////////////////
     {
       super.destroy ( );
     }

     //////////////////////////////////////////////////////////////////////
     // Private methods
     //////////////////////////////////////////////////////////////////////

     private static void  showTable (
       PrintStream  out,
       double       annualSavings,
       double       desiredIncome,
       double       retirementInterestRate,
       double       inflationRate,
       double       taxRate,
       double       investmentInterestRate,
       double       investmentYears )
     //////////////////////////////////////////////////////////////////////
     {
       int  yearsWorking = ( int ) investmentYears;

       if ( ( yearsWorking < 0 ) || ( yearsWorking > 100 ) ) return;

       out.println ( "<TABLE BORDER CELLPADDING=10>" );

       out.println ( "  <THEAD> Investment Years </THEAD>" );
       out.println ( "  <TH> Year                  </TH>" );
       out.println ( "  <TH> Investment Earnings   </TH>" );
       out.println ( "  <TH> Additional Investment </TH>" );
       out.println ( "  <TH> Total Investment      </TH>" );
       out.println ( "  <TH> Present Value         </TH>" );

       double  totalSavings = 0.0;

       for ( int  i = 1; i < yearsWorking + 1; i++ )
       {
         out.println ( "  <TR ALIGN=RIGHT>" );

         out.println ( "    <TD> " + i + "</TD>" );

         double  interest = investmentInterestRate * totalSavings;

         out.println ( "    <TD> "
           + currencyNumberFormat.format ( interest ) + "</TD>" );

         out.println ( "    <TD> "
           + currencyNumberFormat.format ( annualSavings ) + "</TD>" );

         totalSavings = FinanceLib.futureValueAnnuity (
           annualSavings, investmentInterestRate, ( double ) i );

         out.println ( "    <TD> "
           + currencyNumberFormat.format ( totalSavings ) + "</TD>" );

         double  presentValue = FinanceLib.presentValue (
           totalSavings, inflationRate, ( double ) i );

         out.println ( "    <TD> "
           + currencyNumberFormat.format ( presentValue ) + "</TD>" );

         out.println ( "  </TR>" );
       }

       out.println ( "</TABLE>" );

       out.println ( "<P>" );

       out.println ( "<TABLE BORDER CELLPADDING=10>" );

       out.println ( "  <THEAD> First Ten Retirement Years </THEAD>" );
       out.println ( "  <TH> Year                  </TH>" );
       out.println ( "  <TH> Interest Earned       </TH>" );
       out.println ( "  <TH> Tax on Interest       </TH>" );
       out.println ( "  <TH> Living Income         </TH>" );
       out.println ( "  <TH> Remaining Savings     </TH>" );
       out.println ( "  <TH> Present Value         </TH>" );

       for ( int  i = yearsWorking + 1; i < yearsWorking + 11; i++ )
       {
         out.println ( "  <TR ALIGN=RIGHT>" );

         out.println ( "    <TD> " + i + "</TD>" );

         double  interest = retirementInterestRate * totalSavings;

         out.println ( "    <TD> "
           + currencyNumberFormat.format ( interest ) + "</TD>" );

         double  taxes = -taxRate * interest;

         out.println ( "    <TD> "
           + currencyNumberFormat.format ( taxes    ) + "</TD>" );

         double  livingExpenses = -FinanceLib.futureValue (
           desiredIncome, inflationRate, ( double ) i );

         out.println ( "    <TD> "
           + currencyNumberFormat.format ( livingExpenses ) + "</TD>" );

         totalSavings += interest + taxes + livingExpenses;

         out.println ( "    <TD> "
           + currencyNumberFormat.format ( totalSavings ) + "</TD>" );

         double  presentValue = FinanceLib.presentValue (
           totalSavings, inflationRate, ( double ) i );

         out.println ( "    <TD> "
           + currencyNumberFormat.format ( presentValue ) + "</TD>" );

         out.println ( "  </TR>" );
       }

       out.println ( "</TABLE>" );
     }

     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////
     }
