     //////////////////////////////////////////////////////////////////////
     // The contents of this file are subject to the Mozilla Public License
     // Version 1.1 (the "License"); you may not use this file except in
     // compliance with the License.  You may obtain a copy of the License
     // at http://www.mozilla.org/MPL/
     //
     // Software distributed under the License is distributed on an "AS IS"
     // basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
     // See the License for the specific language governing rights and
     // limitations under the License.
     //
     // The Original Code is "com.orbs.open1.mpl.net.news.NNTPSocket",
     // created 1997-05-10.
     //
     // The Initial Developer of the Original Code is David Wallace Croft
     // (croft@alumni.caltech.edu, http://www.alumni.caltech.edu/~croft/).
     // Portions created by David Wallace Croft are
     // Copyright (C) 1999 David Wallace Croft.  All Rights Reserved.
     //
     // Contributor(s):
     //////////////////////////////////////////////////////////////////////

     package com.orbs.open1.mpl.net.news;

     import java.net.*;
     import java.io.*;
     import java.util.*;

     /*********************************************************************
     * Implements NNTP as given by RFC 977.
     * <P>
     * @author
     *   <A HREF="http://www.alumni.caltech.edu/~croft">David W. Croft</A>
     * @version
     *   1997-05-10
     *********************************************************************/

     public class  NNTPSocket extends Socket {
     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////

     public static final int     NNTP_PORT         = 119;

     public static final String  COMMAND_ARTICLE   = "ARTICLE";
     public static final String  COMMAND_BODY      = "BODY";
     public static final String  COMMAND_GROUP     = "GROUP";
     public static final String  COMMAND_HEAD      = "HEAD";
     public static final String  COMMAND_HELP      = "HELP";
     public static final String  COMMAND_IHAVE     = "IHAVE";
     public static final String  COMMAND_LAST      = "LAST";
     public static final String  COMMAND_LIST      = "LIST";
     public static final String  COMMAND_NEWGROUPS = "NEWGROUPS";
     public static final String  COMMAND_NEWNEWS   = "NEWNEWS";
     public static final String  COMMAND_NEXT      = "NEXT";
     public static final String  COMMAND_POST      = "POST";
     public static final String  COMMAND_QUIT      = "QUIT";
     public static final String  COMMAND_SLAVE     = "SLAVE";
     public static final String  COMMAND_STAT      = "STAT";

     public static final String  CR_LF             = "\r\n";

     private DataInputStream     dataInputStream;
     private PrintStream         printStream;

     private String              response_code;
     private Vector              response_text;
     private boolean             posting_allowed;
     private Vector              newsgroup_Vector;

     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////

     /*********************************************************************
     * Be sure to call command_QUIT() when done.
     *********************************************************************/
     public  NNTPSocket ( String  nntp_server_host ) throws IOException {
     //////////////////////////////////////////////////////////////////////
       super ( nntp_server_host, NNTP_PORT );
       dataInputStream = new DataInputStream ( getInputStream ( ) );
       printStream = new PrintStream ( getOutputStream ( ) );
       response_code = dataInputStream.readLine ( );
       if ( response_code.startsWith ( "200" ) ) {
         posting_allowed = true;
       } else if ( !response_code.startsWith ( "201" ) ) {
         throw new IOException ( "Connection failed:  " + response_code );
       }
     }

     /*********************************************************************
     * Returns the response code for the last command.
     *********************************************************************/
     public String  response_code ( ) { return response_code; }

     /*********************************************************************
     * Returns the response text, a Vector of String, for the last command.
     *********************************************************************/
     public Vector  response_text ( ) { return response_text; }

     /*********************************************************************
     * Returns true if posting to this server is allowed.
     *********************************************************************/
     public boolean  posting_allowed ( ) { return posting_allowed; }

     /*********************************************************************
     * Returns a Vector of Newsgroup.
     *********************************************************************/
     public Vector  newsgroup_Vector ( ) { return newsgroup_Vector; }

     /*********************************************************************
     * The original text is broken into an array of String at the
     * delimiters newline, carriage return, or the CR_LF pair.
     * This is returned as a Vector of String without the delimiters.
     * Synchronizes on text.
     *********************************************************************/
     public static Vector  toStringVector ( String  text ) {
     //////////////////////////////////////////////////////////////////////
       String  line;
       Vector  string_Vector = new Vector ( );
       synchronized ( text ) {
         DataInputStream  textInputStream
           = new DataInputStream ( new StringBufferInputStream ( text ) );
         try {
           while ( ( line = textInputStream.readLine ( ) ) != null ) {
             string_Vector.addElement ( line );
           }
         } catch ( Exception  e ) {
           e.printStackTrace ( );
           throw new RuntimeException ( e.getMessage ( ) );
         }
       }
       return string_Vector;
     }

     /*********************************************************************
     * Replaces lines that start with "." with "..",
     * appends CR_LF to each line, then
     * adds a final line of "." + CR_LF.
     *********************************************************************/
     public static StringBuffer  text_encode ( Vector  string_Vector ) {
     //////////////////////////////////////////////////////////////////////
       StringBuffer  stringBuffer = new StringBuffer ( );
       Enumeration  e = string_Vector.elements ( );
       while ( e.hasMoreElements ( ) ) {
         String  line = ( String ) e.nextElement ( );
         if ( line.startsWith ( "." ) ) stringBuffer.append ( "." );
         stringBuffer.append ( line );
         stringBuffer.append ( CR_LF );
       }
       stringBuffer.append ( "." + CR_LF );
       return stringBuffer;
     }

     /*********************************************************************
     * Replaces lines that start with ".." with ".".
     * Synchronizes on text_Vector.
     *********************************************************************/
     public static void  text_decode ( Vector  text_Vector ) {
     //////////////////////////////////////////////////////////////////////
       synchronized ( text_Vector ) {
         int  size = text_Vector.size ( );
         for ( int  i = 0; i < size; i++ ) {
           String  line
             = text_decode ( ( String ) text_Vector.elementAt ( i ) );
           text_Vector.setElementAt ( line, i );
         }
       }
     }

     /*********************************************************************
     * Replaces starting ".." with ".".
     *********************************************************************/
     public static String  text_decode ( String  line ) {
     //////////////////////////////////////////////////////////////////////
       if ( line.startsWith ( ".." ) ) return line.substring ( 1 );
       return new String ( line );
     }

     /*********************************************************************
     * Loads the response_text from the NNTPSocket InputStream.
     * Decodes it as well.
     *********************************************************************/
     public synchronized Vector  load_response_text ( )
       throws IOException {
     //////////////////////////////////////////////////////////////////////
       response_text = new Vector ( );
       String  line;
int  count = 0;
       while ( ( line = dataInputStream.readLine ( ) ) != null ) {
         if ( line.equals ( "." ) ) break;
         line = text_decode ( line );
         response_text.addElement ( line );
if ( ( ++count % 1000 ) == 0 ) System.out.print ( "." );
       }
System.out.println ( "\n" + count + " newsgroups" );
       return response_text;
     }

     /*********************************************************************
     * Sends command + CR_LF to the server.
     * Returns the response_code.
     *********************************************************************/
     public synchronized String  command ( String  command )
       throws IOException {
     //////////////////////////////////////////////////////////////////////
       printStream.println ( command + CR_LF );
       return ( response_code = dataInputStream.readLine ( ) );
     }

     /*********************************************************************
     *********************************************************************/
     public synchronized String  command_GROUP ( String  group_name )
       throws IOException {
     //////////////////////////////////////////////////////////////////////
       response_code = command ( COMMAND_GROUP + " " + group_name );
       if ( response_code.startsWith ( "211" ) ) {
         StringTokenizer  st = new StringTokenizer ( response_code );
         try {
           st.nextToken ( );
           int  estimated_articles = Integer.parseInt ( st.nextToken ( ) );
           int  first_article = Integer.parseInt ( st.nextToken ( ) );
           int  last_article = Integer.parseInt ( st.nextToken ( ) );
           String  new_group_name = st.nextToken ( );
           if ( newsgroup_Vector != null ) {
             Enumeration  e = newsgroup_Vector.elements ( );
             while ( e.hasMoreElements ( ) ) {
               Newsgroup  newsgroup = ( Newsgroup ) e.nextElement ( );
               if ( newsgroup.group_name ( ).equals ( new_group_name ) ) {
                 newsgroup_Vector.removeElement ( newsgroup );
                 break;
               }
             }
           } else newsgroup_Vector = new Vector ( );
           newsgroup_Vector.insertElementAt ( new Newsgroup (
             estimated_articles, first_article, last_article,
             new_group_name ), 0 );
         } catch ( Exception  e ) { e.printStackTrace ( ); }
       }
       return response_code;
     }

     /*********************************************************************
     * Loads the newsgroup_Vector.  Returns the response_code.
     *********************************************************************/
     public synchronized String  command_LIST ( ) throws IOException {
     //////////////////////////////////////////////////////////////////////
       newsgroup_Vector = new Vector ( );
       response_code = command ( COMMAND_LIST );
       if ( response_code.startsWith ( "2" ) ) {
         try { load_response_text ( ); }
         catch ( IOException  e ) { e.printStackTrace ( ); }
         Enumeration  e = response_text.elements ( );
         while ( e.hasMoreElements ( ) ) {
           String  line = ( String ) e.nextElement ( );
           StringTokenizer  st = new StringTokenizer ( line );
           try {
             while ( st.hasMoreTokens ( ) ) {
               String  group_name = st.nextToken ( );
               int  last_article = Integer.parseInt ( st.nextToken ( ) );
               int  first_article = Integer.parseInt ( st.nextToken ( ) );
               boolean  posting_ok
                 = st.nextToken ( ).equalsIgnoreCase ( "y" );
               newsgroup_Vector.addElement ( new Newsgroup (
                 group_name, last_article, first_article, posting_ok ) );
             }
           } catch ( Exception  ex ) { ex.printStackTrace ( ); }
         }
       }
       return response_code;
     }

     /*********************************************************************
     * Sends COMMAND_QUIT and then closes the NNTPSocket.
     *********************************************************************/
     public synchronized String  command_QUIT ( ) throws IOException {
     //////////////////////////////////////////////////////////////////////
       response_code = command ( COMMAND_QUIT );
       close ( );
       return response_code;
     }

     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////
     }
