     //////////////////////////////////////////////////////////////////////
     // The contents of this file are subject to the Mozilla Public License
     // Version 1.1 (the "License"); you may not use this file except in
     // compliance with the License.  You may obtain a copy of the License
     // at http://www.mozilla.org/MPL/
     //
     // Software distributed under the License is distributed on an "AS IS"
     // basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
     // See the License for the specific language governing rights and
     // limitations under the License.
     //
     // The Original Code is
     // "com.orbs.open1.mpl.util.loop.Looper",
     // created 1999-12-23.
     //
     // The Initial Developer of the Original Code is David Wallace Croft
     // (croft@alumni.caltech.edu, http://www.alumni.caltech.edu/~croft/).
     // Portions created by David Wallace Croft are
     // Copyright (C) 1999 David Wallace Croft.  All Rights Reserved.
     //
     // Contributor(s):
     //////////////////////////////////////////////////////////////////////

     package com.orbs.open1.mpl.util.loop;

     import com.orbs.open1.mpl.lang.ex.ExceptionHandler;

     /*********************************************************************
     * Periodically runs a task in a loop.
     *
     * <P>
     *
     * In Java 1.3, a possible alternative to Looper is java.util.Timer.
     *
     * <P>
     *
     * @see
     *   java.util.Timer
     *
     * @author
     *   <A HREF="http://www.alumni.caltech.edu/~croft">David W. Croft</A>
     * @version
     *   2000-04-27
     *********************************************************************/

     public class  Looper implements Runnable
     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////
     {

     private Loopable          loopable;

     private long              period;

     private ExceptionHandler  exceptionHandler;

     private boolean           doRun;

     private boolean           isRunning;

     private Thread            thread;

     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////

     /*********************************************************************
     * Creates a Looper.
     *
     * <P>
     *
     * Call the start() method after construction to begin looping.
     *
     * <P>
     *
     * @param  loopable
     *
     *   The loop() method of this object will be called periodically.
     *   Looping will stop if this method returns false.
     *
     *   <P>
     *
     * @param  period
     *
     *   The delay between runs in milliseconds.
     *
     *   <P>
     *
     * @param  exceptionHandler
     *
     *   Handles all Exceptions generated by runnable.run().
     *   The handleException() Object argument passed will be the loopable.
     *   If null, the default behavior of printing any Exception to the
     *   standard error output and stopping looping will be used.
     *
     *   <P>
     *********************************************************************/
     public  Looper (
       Loopable          loopable,
       long              period,
       ExceptionHandler  exceptionHandler )
     //////////////////////////////////////////////////////////////////////
     {
       if ( loopable == null )
       {
         throw new IllegalArgumentException ( "null loopable" );
       }

       this.loopable         = loopable;

       this.period           = period;

       this.exceptionHandler = exceptionHandler;
     }

     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////

     public long  getPeriod ( )
     //////////////////////////////////////////////////////////////////////
     {
       return period;
     }

     public synchronized void  setPeriod ( long  period )
     //////////////////////////////////////////////////////////////////////
     {
       this.period = period;
     }

     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////

     /*********************************************************************
     * @return
     *
     *   Returns false if already running.
     *********************************************************************/
     public synchronized boolean  start ( )
     //////////////////////////////////////////////////////////////////////
     {
       if ( isRunning )
       {
         return false;
       }

       isRunning = true;

       doRun = true;

       ( thread = new Thread ( this ) ).start ( );

       return true;
     }

     /*********************************************************************
     * Do not call this method directly.  Do not override.
     *********************************************************************/
     public final void  run ( )
     //////////////////////////////////////////////////////////////////////
     {
       if ( Thread.currentThread ( ) != thread )
       {
         // perhaps this should be in a separate class for privacy

         throw new IllegalStateException ( );
       }

       while ( doRun )
       {
         try
         {
           if ( !loopable.loop ( ) )
           {
             break;
           }
         }
         catch ( Exception  ex )
         {
           if ( ( exceptionHandler == null )
             || !exceptionHandler.handleException ( ex, loopable ) )
           {
             ex.printStackTrace ( );

             break;
           }
         }

         try
         {
           thread.sleep ( period );
         }
         catch ( InterruptedException  ex )
         {
//         break;
         }
       }

       thread = null;

       isRunning = false;
     }

     /*********************************************************************
     * Requests thread termination at the end of this loop.
     *
     * <P>
     *
     * Interrupts the internal thread and sets the "continue looping" flag
     * to false.
     *
     * If the internal thread is interrupted for any reason,
     * regardless of whether it was interrupted by calling this method,
     * the internal thread will terminate at the end of the current loop.
     *
     * <P>
     *
     * The loop can be terminated sooner if the runnable.run() method is
     * implemented to check for thread interruption or if this method is
     * overridden.  For example, one might override this method to force
     * quick termination by first calling this superclass method and then
     * closing any sockets or other open resources necessary for operation
     * in the subclass method.
     *********************************************************************/
     public synchronized void  stop ( )
     //////////////////////////////////////////////////////////////////////
     {
       doRun = false;

       try { thread.interrupt ( ); } catch ( NullPointerException  ex ) { }

       loopable.abort ( );
     }

     //////////////////////////////////////////////////////////////////////
     //////////////////////////////////////////////////////////////////////
     }
